/**
 *  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License"). You may not use this file except in compliance
 *  with the License. A copy of the License is located at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  or in the 'license' file accompanying this file. This file is distributed on an 'AS IS' BASIS, WITHOUT WARRANTIES
 *  OR CONDITIONS OF ANY KIND, express or implied. See the License for the specific language governing permissions
 *  and limitations under the License.
 */
import { CfnAgent } from 'aws-cdk-lib/aws-bedrock';
import { IBucket, Location } from 'aws-cdk-lib/aws-s3';
/******************************************************************************
 *                       API SCHEMA CLASS
 *****************************************************************************/
/**
 * Represents the concept of an API Schema for a Bedrock Agent Action Group.
 */
export declare abstract class ApiSchema {
    /**
     * Creates an API Schema from a local file.
     * @param path - the path to the local file containing the OpenAPI schema for the action group
     */
    static fromLocalAsset(path: string): InlineApiSchema;
    /**
     * Creates an API Schema from an inline string.
     * @param schema - the JSON or YAML payload defining the OpenAPI schema for the action group
     */
    static fromInline(schema: string): InlineApiSchema;
    /**
     * Creates an API Schema from an S3 File
     * @param bucket - the bucket containing the local file containing the OpenAPI schema for the action group
     * @param objectKey - object key in the bucket
     */
    static fromS3File(bucket: IBucket, objectKey: string): S3ApiSchema;
    readonly s3File?: Location;
    readonly inlineSchema?: string;
    /**
     * Constructor accessible only to extending classes.
     */
    protected constructor(s3File?: Location, inlineSchema?: string);
    /**
     * Format as CFN properties
     *
     * @internal This is an internal core function and should not be called directly.
     */
    abstract _render(): CfnAgent.APISchemaProperty;
}
export declare class InlineApiSchema extends ApiSchema {
    private readonly schema;
    constructor(schema: string);
    /**
     * @internal This is an internal core function and should not be called directly.
     */
    _render(): CfnAgent.APISchemaProperty;
}
/**
 * Class to define an API Schema from an S3 object.
 */
export declare class S3ApiSchema extends ApiSchema {
    private readonly location;
    constructor(location: Location);
    /**
     * @internal This is an internal core function and should not be called directly.
     */
    _render(): CfnAgent.APISchemaProperty;
}
